# AUTOGENERATED! DO NOT EDIT! File to edit: 02_processing.ipynb (unless otherwise specified).

__all__ = ['eyetrack_stim_inten', 'saccade_distances', 'smooth_eye_position', 'process_sta_batch', 'staEst_fromBins',
           'process_sta_batch_large', 'cross_correlation', 'corrcoef', 'flatten_corrcoef', 'stimulus_ensemble',
           'process_nonlinearity', 'activity_histogram', 'cross_distances', 'cross_distances_sta', 'paired_distances',
           'paired_distances_sta', 'direction_selectivity', 'wave_direction_selectivity', 'peri_saccadic_response']

# Cell
from functools import partial
import numpy as np
from sklearn.decomposition import PCA
from sklearn import cluster
import scipy.ndimage as ndimage
import scipy.signal as signal
import scipy as sp
from cmath import *
import itertools
import random

from .core import *
from .utils import *
from .modelling import *
from .leddome import *

# Cell
def eyetrack_stim_inten(stim_inten, eye_track,
                        upsampling=2,
                        eye_calib=[[94 ,8], [ 18, 59]],
                        box_w=None, box_h=None, stim_axis="x"):
    """
    From stimulus data and eye tracking, returns a corrected and upsampled stimulus data.
    Calibration corresponds to the width and height of the stimulus screen, in
    terms of pixels in the eye video: [[WIDTH_x, WIDTH_y], [HEIGHT_x, HEIGHT_y]]

    params:
        - stim_inten: Stimulus intensity matrix of shape (t, y, x), or (t, x) or (t, y) depending on stim_axis
        - eye_track: Eye tracking data of shape (t, x_pos, y_pos, ...)
        - upsampling: Factor for the upsampling (2 will multiply by 2 number of box in width and height)
        - eye_calib: Calibration matrix of shape (2,2)
        - box_w: Width of a block in pixel (40px in case of a 32 box in width of a checkerboard on a 1280px width)
        - box_h: Height of a block in pixel. Both box_x and box_h are calculated from a 1280x720 screen if None
        - stim_axis: Specify which direction to shift in case of stim shape different than (t, y, x)

    return:
        - Upsampled and shift corrected stimulus intensity
    """
    eye_x, eye_y = eye_track[:,0], eye_track[:,1]
    shape_y, shape_x = 1, 1
    if len(stim_inten.shape)==2:
        if stim_axis=="x":
            shape_x = stim_inten.shape[1]
        elif stim_axis=="y":
            shape_y = stim_inten.shape[1]
    elif len(stim_inten.shape)==3:
        shape_y = stim_inten.shape[1]
        shape_x = stim_inten.shape[2]
    if box_w is None:
        box_w = 1280//shape_x
    if box_h is None:
        box_h = 720//shape_y

    if shape_y>1 and shape_x>1:
        box_w, box_h = int(box_w/upsampling), int(box_h/upsampling)
    elif shape_x > 1:
        box_w, box_h = int(box_w/upsampling), box_h
    elif shape_y > 1:
        box_w, box_h = box_w                , int(box_h/upsampling)

    eye_transfo_f = _eye_to_stim_f(eye_calib=eye_calib,
                                  box_width=box_w,
                                  box_height=box_h)

    if shape_y>1 and shape_x>1:
        stim_inten = stim_inten.repeat(upsampling,axis=1).repeat(upsampling,axis=2)
    else:
        stim_inten = stim_inten.repeat(upsampling,axis=1)

    xpos_avg = np.mean(eye_x)
    ypos_avg = np.mean(eye_y)
    mean_stim_inten = int((np.max(stim_inten)+np.min(stim_inten))/2)
    #After getting the shift of the matrix to apply, we roll the matrix instead of extending it to the shifts
    #This seems strange, but from the cell point of view, that is potentially looking at no stimulus,
    # the response it gives are uncorrelated with the stimulus, and so shouldn't impact further analysis
    # Advantage is that it keeps the data small enough, without loosing regions of the stimulus.
    for i in range(len(stim_inten)):
        x_eyeShift = eye_x[i]-xpos_avg
        y_eyeShift = eye_y[i]-ypos_avg

        stim_shift_x, stim_shift_y = eye_transfo_f(x_eyeShift=x_eyeShift,
                                                   y_eyeShift=y_eyeShift)
        if shape_y>1 and shape_x>1:
            rolled_stim = np.roll(stim_inten[i],stim_shift_y,axis=0)
            rolled_stim = np.roll(rolled_stim  ,stim_shift_x,axis=1)
        else:
            if stim_axis=="x":
                rolled_stim = np.roll(stim_inten[i],stim_shift_x,axis=0)
            else:
                rolled_stim = np.roll(stim_inten[i],stim_shift_y,axis=0)
        stim_inten[i] = rolled_stim

    return stim_inten

def saccade_distances(eye_track):
    """
    Create a mask for the eye position timeserie that indicate how far was the last saccade.
    The eye positions need to be smoothed with smooth_eye_position.

    params:
        - eye_track: Eye tracking data of shape (t, x_pos, y_pos, ...)

    return:
        - Distance to last saccade array
    """

    x_pos, y_pos = eye_track[:,0], eye_track[:,1]
    saccade_pos = np.where((x_pos[1:] != x_pos[:-1]) & (y_pos[1:] != y_pos[:-1]))[0] + 1

    len_chunks = [saccade_pos[0]]+list(saccade_pos[1:]-saccade_pos[:-1])
    len_chunks.append(len(x_pos) - saccade_pos[-1])

    saccade_mask = []
    for len_chunk in len_chunks:
        saccade_mask.extend(list(range(len_chunk)))

    return np.array(saccade_mask)

def smooth_eye_position(eye_track, threshold=2):
    """
    Smooth the eye positions. Uses clustering of eye position epochs to smooth.
    Clustering is made with sklearn.cluster.dbscan.

    params:
        - eye_track: Eye tracking data of shape (t, x_pos, y_pos, ...)
        - threshold: Distance threshold in the position clustering

    return:
        - Distance to last saccade array
    """
    x_pos, y_pos = eye_position[:,0], eye_position[:,1]

    X = np.stack((x_pos,y_pos, np.linspace(0, len(x_pos)/2, len(x_pos)))).T
    clusters = cluster.dbscan(X, eps=threshold, min_samples=3, metric='minkowski', p=2)

    move_events = np.where(clusters[1][1:] > clusters[1][:-1])[0] + 1

    len_chunks = [move_events[0]]+list(move_events[1:]-move_events[:-1])
    len_chunks.append(len(x_pos) - move_events[-1])

    eye_x_positions = np.split(x_pos, move_events)
    eye_y_positions = np.split(y_pos, move_events)

    mean_x_pos = np.array(list(map(np.mean, eye_x_positions)))
    mean_y_pos = np.array(list(map(np.mean, eye_y_positions)))

    x_pos_smooth = np.concatenate([[x_pos]*len_chunk for x_pos,len_chunk in zip(mean_x_pos, len_chunks)])
    y_pos_smooth = np.concatenate([[y_pos]*len_chunk for y_pos,len_chunk in zip(mean_y_pos, len_chunks)])

    return np.stack((x_pos_smooth, y_pos_smooth)).T

def _eye_to_stim_f(eye_calib, box_width, box_height):
    """
    Initialise stimulus shift function parameters and returns a partial function.

    params:
        - eye_calib: Calibration matrix of shape (2,2)
        - box_width: Width of the elementary block in pixel
        - box_height: Width of the elementary block in pixel

    return:
        - Partial function to obtain stimulus shift from eye positions
    """
    eye_to_stim     = np.linalg.inv(eye_calib)
    box_dim         = np.array([1280/(box_width), 720/(box_height)])
    return partial(_linear_transform, box_dim=box_dim, transfo_matrix=eye_to_stim)

def _linear_transform(box_dim, transfo_matrix, x_eyeShift, y_eyeShift):
    """
    Computes the shift of the stimulus from eye position shifts.

    params:
        - box_dim: Size in pixel of the box [width, height]
        - transfo_matrix: Inverse matrix of the calibration matrix described in eyetrack_stim_inten
        - x_eyeShift: Eye shift in x
        - y_eyeShift: Eye shift in y

    return:
        - Stimulus shift tuple (delta_x, delta_y)
    """
    transform_coord = np.dot(transfo_matrix, np.array([x_eyeShift, y_eyeShift]).T)
    stim_vec        = np.round(transform_coord * box_dim).astype(int)
    return stim_vec[0], -stim_vec[1]

# Cell
def process_sta_batch(stim_inten, spike_counts, Hw=30, Fw=2, return_pval=False, normalisation="abs"):
    """
    Computes the STA and associated pvalues in parallel for a batch of cells.

    params:
        - stim_inten: stimulus intensity matrix of shape (t, ...)
        - spike_counts: cells activity matrix of shape (t, n_cell)
        - Hw: Lenght in frames of the history window, including the 0 timepoint
        - Fw: Lenght in frames of the forward window
        - return_pval: Flag to signal whether or not to return the pvalues
        - normalisation: Normalization applied to the STA. One of ["abs", "L2", None]

    return:
        - stas of shape (n_cell, Hw+Fw, ...)
        - stas and pvalues if return_pval=True, both of shape (n_cell, Hw+Fw, ...)
    """
    assert normalisation in ["abs", "L2", None], "normalisation must be one of ['abs', 'L2', None]"
    #Preparing the stimulus
    orig_shape = stim_inten.shape
    stim_inten = stim_inten_norm(stim_inten)
    sum_spikes = np.sum(spike_counts, axis=0)
    len_stim = len(stim_inten)

    #We just have to calculate one STA over the whole record
    stim_inten   = np.reshape(stim_inten, (len(stim_inten),-1))
    stim_inten   = np.transpose(stim_inten)
    allCells_sta = staEst_fromBins(stim_inten, spike_counts, Hw, Fw=Fw)

    if len(orig_shape)==3:
        allCells_sta = allCells_sta.reshape((len(allCells_sta),Hw+Fw, orig_shape[-2], orig_shape[-1]))
    elif len(orig_shape)==2:
        allCells_sta = allCells_sta.reshape((len(allCells_sta),Hw+Fw,-1))
    else:
        allCells_sta = allCells_sta.reshape((len(allCells_sta),Hw+Fw))

    if allCells_sta.shape[0]==1: #Only one cell, but we need to keep the axis
        allCells_sta = np.squeeze(allCells_sta)
        allCells_sta = np.expand_dims(allCells_sta, axis=0)
    else:
        allCells_sta = np.squeeze(allCells_sta)

    if return_pval:
        p_values = np.empty(allCells_sta.shape)
    for k, cell_sta in enumerate(allCells_sta): #Easy way to do normalization for each cell that works for all possible shapes
        if return_pval:
            z_scores    = cell_sta/ np.sqrt(1/sum_spikes[k]) #Standard score is calculated as (x-mean)/std
            p_values[k] = sp.stats.norm.sf(abs(z_scores))*2

        if normalisation is None:
            continue
        elif normalisation=="abs":
            allCells_sta[k] = np.nan_to_num(cell_sta/np.max(np.abs(cell_sta)))
        elif normalisation=="L2":
            allCells_sta[k] = np.nan_to_num(cell_sta/np.sqrt(np.sum(np.power(cell_sta, 2))))

    if return_pval:
        return allCells_sta, p_values
    else:
        return allCells_sta

def staEst_fromBins(stim, spike_counts, Hw, Fw=0):
    """
    Matrix mutliplication to compute the STA. Use the wrapper process_sta_batch to avoid bugs.

    params:
        - stim_inten: stimulus intensity matrix of shape (flattened_frame, t)
        - spike_counts: cells activity matrix of shape (t, n_cell)
        - Hw: Lenght in frames of the history window, including the 0 timepoint
        - Fw: Lenght in frames of the forward window

    return:
        - STA of shape (n_cell, Hw+Fw, flattened_frame)
    """
    spike_counts[:Hw] = 0

    spike_counts = np.nan_to_num(spike_counts / np.sum(spike_counts,axis=0))
    spike_counts = spike_counts - np.mean(spike_counts, axis=0) #Center to 0 the spike counts to include "inhibitory" stimulus
    sta = np.zeros((Hw+Fw, stim.shape[0], spike_counts.shape[-1]))
    for i in range(Hw): #Does the dot product of cell activity and frame intensity matrix
        sta[(Hw-1-i),:,:] = np.dot(stim, spike_counts)
        spike_counts      = np.roll(spike_counts, -1, axis=0) #And shift the activity to compute the next frame
    spike_counts = np.roll(spike_counts, Hw, axis=0)
    if Fw != 0:
        spike_counts[-Fw:] = 0
    for i in range(Fw): #Same thing for Fw
        spike_counts  = np.roll(spike_counts, 1, axis=0)
        sta[Hw+i,:,:] = np.dot(stim, spike_counts)
    spike_counts = np.roll(spike_counts, -Fw, axis=0)
    return np.transpose(sta, (2,0,1))

def process_sta_batch_large(stim_inten, spike_counts, Hw=30, Fw=2, return_pval=False, normalisation="abs", bs=1000):
    """
    Computes the STA and associated pvalues in parallel for a batch of cells, for a large stimulus.

    params:
        - stim_inten: stimulus intensity matrix of shape (t, ...)
        - spike_counts: cells activity matrix of shape (t, n_cell)
        - Hw: Lenght in frames of the history window, including the 0 timepoint
        - Fw: Lenght in frames of the forward window
        - return_pval: Flag to signal whether or not to return the pvalues
        - normalisation: Normalization applied to the STA. One of ["abs", "L2", None]
        - bs: batch size to compute partial STA

    return:
        - stas of shape (n_cell, Hw+Fw, ...)
        - stas and pvalues if return_pval=True, both of shape (n_cell, Hw+Fw, ...)
    """
    orig_shape = stim_inten.shape
    n_spatial_dim = orig_shape[1]*orig_shape[2]

    sum_spikes = np.sum(spike_counts, axis=0)
    len_stim = len(stim_inten)
    allCells_sta = np.zeros((n_spatial_dim, spike_counts.shape[1], Hw+Fw))
    stim_inten = stim_inten.reshape((len_stim,-1))
    print("Computing the STA part by part:")
    for i, batch_pos in enumerate(range(0, n_spatial_dim, bs)):
        print(str(round(100*batch_pos/n_spatial_dim,2))+"%      ", end="\n", flush=True)
        #Computing STA on partial portions of the screen sequentially
        stim_part = stim_inten_norm(stim_inten[:, batch_pos:batch_pos+bs]).T
        sub_sta = staEst_fromBins(stim_part, spike_counts, Hw, Fw=Fw)
        allCells_sta[batch_pos:batch_pos+bs] = np.transpose(sub_sta, (2,0,1))#(ncell,Hw,stim_len) to (stim_len,ncell,Hw)
    allCells_sta = np.transpose(allCells_sta, (1,2,0))
    print("100%      ")

    if len(orig_shape)==3:
        allCells_sta = allCells_sta.reshape((len(allCells_sta),Hw+Fw, orig_shape[-2], orig_shape[-1]))
    elif len(orig_shape)==2:
        allCells_sta = allCells_sta.reshape((len(allCells_sta),Hw+Fw,-1))
    else:
        allCells_sta = allCells_sta.reshape((len(allCells_sta),Hw+Fw))

    if allCells_sta.shape[0]==1: #Only one cell, but we need to keep the axis
        allCells_sta = np.squeeze(allCells_sta)
        allCells_sta = np.expand_dims(allCells_sta, axis=0)
    else:
        allCells_sta = np.squeeze(allCells_sta)

    if return_pval:
        p_values = np.empty(allCells_sta.shape)
    for k, cell_sta in enumerate(allCells_sta): #Easy way to do normalization for each cell that works for all possible shapes
        if return_pval:
            z_scores    = cell_sta/ np.sqrt(1/sum_spikes[k]) #Standard score is calculated as (x-mean)/std
            p_values[k] = sp.stats.norm.sf(abs(z_scores))*2

        if normalisation is None:
            continue
        elif normalisation=="abs":
            allCells_sta[k] = np.nan_to_num(cell_sta/np.max(np.abs(cell_sta)))
        elif normalisation=="L2":
            allCells_sta[k] = np.nan_to_num(cell_sta/np.sqrt(np.sum(np.power(cell_sta, 2))))

    if return_pval:
        return allCells_sta, p_values
    else:
        return allCells_sta

# Cell
def cross_correlation(spike_counts, tail_len=100):
    """
    From calculate the cross correlation of the cells over a time window.

    params:
        - spike_counts of shape (t, n_cell)
        - tail_len: time correlation window size

    return:
        - cross correlation between the cells
    """
    n_dpoints, n_cell = spike_counts.shape
    corr_arr = np.zeros((n_cell,n_cell,tail_len+1))
    spike_counts = (spike_counts / np.max(spike_counts, axis=0)) #Independant normalization of the cells
    spike_counts_edged = np.concatenate((np.zeros((tail_len,n_cell)),
                    spike_counts,
                    np.zeros((tail_len,n_cell)))) #Creating an array with 0 tails on both side to use the valid mode
                                                  # of numpy.correlate
    for i in range(n_cell):
        for j in range(i, n_cell):
            corr_arr[i,j] = np.correlate(spike_counts_edged[:,i],
                                         spike_counts[:,j],
                                         mode="valid")
            corr_arr[j,i] = corr_arr[i,j]
    return corr_arr/n_dpoints

# Cell
def corrcoef(spike_counts):
    """
    Computes correlation coefficient between the cells

    params:
        - spike_counts: Cells activity of shape (t, n_cell)

    return:
        - Correlation matrix
    """
    return np.corrcoef(spike_counts.T)

def flatten_corrcoef(corrcoef_matrix):
    """
    Retrieves a one dimensional array of the cells correlations

    params:
        - corrcoef_matrix: Correlation matrix from `corrcoef`

    return:
        - flattened correlation matrix"""
    shp = corrcoef_matrix.shape
    return np.array([corrcoef_matrix[i,j] for i in range(shp[0]) for j in range(i+1, shp[0])])

# Cell
def stimulus_ensemble(stim_inten, Hw=30, x=0, y=0, w=None, h=None):
    """
    Generate the stimulus ensemble used to compute the nonlinearity

    params:
        - stim_inten: stimulus intensity matrix of shape (t, ...)
        - Hw: Lenght in frames of the history window, including the 0 timepoint
        - x: Left position of the window where to get the ensemble from
        - y: Up position of the window where to get the ensemble from
        - w: Width of the window where to get the ensemble from. If None, is set to stim_inten.shape[2]
        - h: Height of the window where to get the ensemble from. If None, is set to stim_inten.shape[1]

    return:
        - Flatten stimulus ensemble of size (len(stim_inten)-(Hw-1), w*h*Hw). To obtain the corresponding cell activity,
        slice it like so: slice(Hw-1, None)
    """
    stim_inten = stim_inten_norm(stim_inten)
    if len(stim_inten.shape) == 1:
        stim_inten = stim_inten[..., np.newaxis, np.newaxis]
    elif len(stim_inten.shape) == 2:
        stim_inten = stim_inten[..., np.newaxis]

    if w is None:
        w = stim_inten.shape[2]
    if h is None:
        h = stim_inten.shape[1]
    xmin, xmax = max(0,x-w), min(stim_inten.shape[2], x+w+1)
    ymin, ymax = max(0,y-h), min(stim_inten.shape[1], y+h+1)
    dtype = stim_inten.dtype
    if np.all(np.in1d(stim_inten, [-1,0,1])):
        dtype = "int8"
    stim_ensmbl = np.zeros((len(stim_inten)-(Hw-1), (xmax-xmin)*(ymax-ymin)*Hw), dtype=dtype)
    for i in range(0, len(stim_inten)-(Hw-1)):
        flat_stim         = np.ndarray.flatten(stim_inten[i:i+Hw,
                                                          ymin:ymax,
                                                          xmin:xmax])
        stim_ensmbl[i] = flat_stim
    return stim_ensmbl

def process_nonlinearity(stim_inten, spike_counts, bins, stas, p_norm=2):
    """
    Computes the nonlinearity of a single cell. The STA of the cell is in L2 normalization, which
    should restrict the histogram values.

    params:
        - stim_inten: stimulus intensity in shape (t, y, x)
        - spike_counts: cells activity in shape (t, n_cell)
        - bins: bins in which the transformed stimuli ensembles are set. (usually between -6 and 6)
        - stas:  The STAs to convolve with stim_inten in shape (n_cell, Hw, ...)
        - p_norm: Power for the normalization. https://en.wikipedia.org/wiki/Norm_(mathematics)#p-norm :
            1 -> can compare nonlinearites of stimuli with different dimensionality
            2 -> Common L2 normalization for STAs

    return:
        - nonlinearity of the cell.
    """
    assert len(stim_inten)==len(spike_counts)
    stim_inten    = stim_inten_norm(np.array(stim_inten))

    nonlins = np.empty((len(stas), len(bins)-1))

    compute_with_dotprod = False
    try:
        stim_ensemble = stimulus_ensemble(stim_inten, Hw=stas.shape[1])
        compute_with_dotprod = True
    except MemoryError as err:
        print("Not enough memory to generate the stimulus ensemble, "+
              "computing nonlinearity with cross-correlation (will be slower)")

    #In the case of calcium imaging, we have probabilities, not spike counts. Need to make it integers
    # The discretisation of the calcium imaging is done here globally (for all cells together)
    # If it's not what you want, either do the discretisation inside the loop bellow, or discretise the S_matrix
    # before passing it to this function

    if np.max(spike_counts)<1:
        mask         = np.where(spike_counts > 0)
        nonzero_min  = np.min(spike_counts[mask])
        discretized  = spike_counts/nonzero_min
        spike_counts = ((10*discretized)/(np.max(discretized))).astype(int)

    spike_counts  = np.array(spike_counts)[stas.shape[1]-1:].astype(int)

    for i, (sta, sp_count) in enumerate(zip(stas, spike_counts.T)):
        sta /= np.power(np.sum(np.power(np.abs(sta), p_norm)), 1/p_norm) # p-norm

        if not sp_count.any(): #No spikes
            continue

        if compute_with_dotprod:
            #This one is faster, but requires the stim_ensemble to fit the computer memory
            filtered_stim = stim_ensemble@sta.reshape(-1)
        else:
            filtered_stim = np.squeeze(sp.signal.correlate(stim_inten, sta, mode="valid"))

        filtered_sptrigg = np.repeat(filtered_stim, sp_count)

        hist_all   = np.histogram(filtered_stim, bins=bins)[0]
        hist_trigg = np.histogram(filtered_sptrigg, bins=bins)[0]

        nonlin = hist_trigg/hist_all

        if np.count_nonzero(~np.isnan(nonlin))<2: #Less than two values in the nonlin, cannot fill the gaps
            nonlin = np.nan_to_num(nonlin)
        else:
            nonlin = np.nan_to_num(fill_nan(nonlin))

        nonlins[i] = nonlin

    return nonlins


# Cell
def activity_histogram(spike_counts):
    """
    Retrieve an histogram of the individual cells activity.

    params:
        - spike_counts: cells activity matrix of shape (t, n_cell)

    return:
        - Cells activity histogram
    """
    flat_spikes = spike_counts.reshape(-1)
    flat_cell = np.array([[i]*spike_counts.shape[0] for i in range(spike_counts.shape[1])]).reshape(-1)
    hist = np.histogram2d(flat_spikes, flat_cell, bins=[100,spike_counts.shape[1]])[0] / spike_counts.shape[0]
    return hist

# Cell
def cross_distances(masks):
    """
    Computes cross distances from the center of mass of a list of mask.

    params:
        - masks: cells mask of shape (n_mask, y, x)

    return:
        - cross distances matrix
    """
    center_mass = np.array([ndimage.measurements.center_of_mass(mask) for mask in masks])
    cross_distances = np.zeros((len(masks),len(masks)))
    for i in range(len(masks)):
        for j in range(i,len(masks)):
            cross_distances[i,j] = np.linalg.norm(center_mass[i]-center_mass[j])
            cross_distances[j,i] = cross_distances[i,j]
    return cross_distances

def cross_distances_sta(fits, sta_shape, f):
    """
    Computes cross distances between STAs

    params:
        - fits: fits of STA to compare
        - sta_shape: shape of the STA
        - f: function used to compute the stas from the fits parameters

    return:
        - cross distances matrix
    """
    sta_masks = np.array([img_2d_fit(sta_shape, fit, f) for fit in fits])
    for i,sta_mask in enumerate(sta_masks):
        if abs(np.min(sta_mask)) > np.max(sta_mask):
            sta_masks[i] = sta_mask < -.5
        else:
            sta_masks[i] = sta_mask > .5
    return cross_distances(sta_masks)

def paired_distances(masks_1, masks_2):
    """
    Compute the distance between two masks

    params:
        - masks_1: Mask of the first cell
        - masks_2: Mask of the second cell

    return:
        - distance between the two masks
    """
    center_mass_1 = np.array([ndimage.measurements.center_of_mass(mask) for mask in masks_1])
    center_mass_2 = np.array([ndimage.measurements.center_of_mass(mask) for mask in masks_2])
    paired_distances = np.zeros(len(masks_1))
    for i, (center_1, center_2) in enumerate(zip(masks_1, masks_2)):
        paired_distances[i] = np.linalg.norm(center_1-center_2)
    return paired_distances

def paired_distances_sta(sta_fits_1, sta_fits_2, sta_shape, f):
    """
    Compute the distance between two STAs.

    params:
        - sta_fits_1: fits of STA of the first cell
        - sta_fits_2: fits of STA of the second cell
        - sta_shape: shape of the STA
        - f: function used to compute the stas from the fits parameters

    return:
        - distance between the two STAs
    """
    sta_masks_1 = np.array([img_2d_fit(sta_shape, fit, f) for fit in sta_fits_1])
    for i,sta_mask in enumerate(sta_masks_1):
        if abs(np.min(sta_mask)) > np.max(sta_mask):
            sta_masks_1[i] = sta_mask < -.5
        else:
            sta_masks_1[i] = sta_mask > .5
    sta_masks_2 = np.array([img_2d_fit(sta_shape, fit, f) for fit in sta_fits_2])
    for i,sta_mask in enumerate(sta_masks_2):
        if abs(np.min(sta_mask)) > np.max(sta_mask):
            sta_masks_2[i] = sta_mask < -.5
        else:
            sta_masks_2[i] = sta_mask > .5
    return paired_distances(sta_masks_1, sta_masks_2)

# Cell
def direction_selectivity(grouped_spikes_d, n_bootstrap=1000):
    """
    Compute the direction selectivity index of cells in the given dict containing for each condition as
    the keys, an array of shape (n_angle, n_repeat, trial_len, n_cell). Such dictionnary can be obtained
    by using `utils.group_direction_response`.

    params:
        - grouped_spikes_d: Results of the group_direction_response of shape (n_angle, n_repeat, t, n_cell)
        - n_bootstrap: Number of bootstrap iteration to calculate the p-value

    return:
        - A dictionnary with a key for each condition retrieving a tuple containing list of the cells
        (spike sum, direction pref, DS idx, orientation pref, OS idx, pval DS, pval OS)
    """

    res_d = {}
    for cond, sp_count in grouped_spikes_d.items():
        n_angle = sp_count.shape[0]
        sum_rep_spike = np.sum(sp_count, axis=(1,2)).T

        x         = np.linspace(0, (n_angle-1)/4*np.pi, num=n_angle)

        #Direction selectivity
        vect_dir  = np.exp(x*1j)#np.array([np.cos(x) + np.sin(x)*1j])
        dir_pref  = np.nan_to_num((vect_dir * sum_rep_spike).sum(axis=1) / sum_rep_spike.sum(axis=1))
        ds_idx    = abs(dir_pref)

        #Orientation selectivity
        vect_ori  = np.exp(x*1j*2)#np.concatenate((vect_dir[:,:n_angle//2], vect_dir[:,:n_angle//2]), axis=1)
        ori_pref  = np.nan_to_num((vect_ori * sum_rep_spike).sum(axis=1) / sum_rep_spike.sum(axis=1))
        ori_idx   = abs(ori_pref)

        #Generating direction and orientation index from shuffled trials
        axtup_l = list(itertools.product(range(sp_count.shape[0]), range(sp_count.shape[1])))
        random.seed(1)
        axtup_l_shuffled = axtup_l.copy()
        rand_ori_idx_l = np.empty((n_bootstrap, sp_count.shape[3]))
        rand_dir_idx_l = np.empty((n_bootstrap, sp_count.shape[3]))
        for i in range(n_bootstrap):
            random.shuffle(axtup_l_shuffled)
            shuffled_sp_count = np.empty(sp_count.shape)
            for axtup, axtup_shuff in zip(axtup_l, axtup_l_shuffled):
                shuffled_sp_count[axtup] = sp_count[axtup_shuff]
            rand_sum_rep_spike      = np.sum(shuffled_sp_count, axis=(1,2)).T
            rand_dir_pref     = np.nan_to_num((vect_dir * rand_sum_rep_spike).sum(axis=1) / rand_sum_rep_spike.sum(axis=1))
            rand_dir_idx_l[i] = abs(rand_dir_pref)

            rand_ori_pref     = np.nan_to_num((vect_ori * rand_sum_rep_spike).sum(axis=1) / rand_sum_rep_spike.sum(axis=1))
            rand_ori_idx_l[i] = abs(rand_ori_pref)

        #Same calculation of pval as in Baden et al 2016
        p_val_dir = 1 - (np.sum(rand_dir_idx_l<ds_idx, axis=0)/n_bootstrap)
        p_val_ori = 1 - (np.sum(rand_ori_idx_l<ori_idx, axis=0)/n_bootstrap)

        #Finally we have to transform the orientation selectivity vectors to put them back in their
        # original orientation, by divinding the phase of the vector by two
        tau = np.pi*2
        polar_ori_pref = np.array(list((map(polar, ori_pref))))
        polar_ori_pref[:,1] = ((polar_ori_pref[:,1]+tau)%tau)/2 #Convert to positive radian angle and divide by two
        ori_pref = np.array([rect(pol[0], pol[1]) for pol in polar_ori_pref])

        res_d[cond] = (sum_rep_spike, dir_pref, ds_idx, ori_pref, ori_idx, p_val_dir, p_val_ori)

    return res_d

# Cell
def wave_direction_selectivity(wave_array, spike_counts, moving_distance_th=1, looming_distance_th=.3, n_bootstrap=1000):
    """
    Computes the direction, orientation and looming/shrinking indexes of the cells in response to the wave stimulus (in LED dome).

    params:
        - wave_array: The indexes of the waves from the record master.
        - spike_counts: The cells response to the waves from the record master.
        - moving_distance_th: Distance threshold in radians above which a stimulus can be considered for the OS and DS.
        - looming_distance_th: Distance threshold in radians bellow which a stimulus can be considered for the looming/shrinking response
        - n_bootstrap: Number of repetition in the bootstraping method to compute the pvalues

    return:
        - A tuple containing lists of the cells response, in order:
            * [0] summed_responses: Summed response of a cell to each wave condition
            * [1] dir_pref_l: Direction preference vector
            * [2] dir_idx_l : Direction indexes
            * [3] dir_pval_l: Direction p_values
            * [4] ori_pref_l: Orientation preference vector
            * [5] ori_idx_l : Orientation indexes
            * [6] ori_pval_l: Orientation p_values
            * [7] loom_idx_l : Looming indexes
            * [8] loom_pval_l: Looming p_values
            * [9] stas_position_l: Position (theta, phi) of the cells receptive fields with the water stimulus
            * [10] waves_position_l: Relative positions of the waves to the cells STA.
    """
    tau = np.pi*2

    indexes, order = np.unique(wave_array, return_index=True)
    epoch_sequence = indexes[1:][np.argsort(order[1:])]
    wave_inten     = build_wave_stimulus_array(epoch_sequence)
    stas_wave      = process_sta_batch(wave_inten, spike_counts, Hw=1, Fw=0, return_pval=False)
    #Hw of 1 because there is a high temporal correlation in the stimulus, so that's enough to find the RF

    summed_responses = np.zeros((100, spike_counts.shape[1]))

    for i in indexes[1:]: #Iterate from 0 to n_wave-1
        where = np.where(wave_array==i)[0]
        summed_responses[i,:] = np.sum(spike_counts[where,:], axis=0)

    summed_responses = summed_responses.T

    dome_positions = get_dome_positions(mode="spherical")

    ori_pref_l, dir_pref_l              = [], []
    ori_idx_l, dir_idx_l, loom_idx_l    = [], [], []
    ori_pval_l, dir_pval_l, loom_pval_l = [], [], []
    stas_position_l, waves_position_l   = [], []

    for sta, cell_responses  in zip(stas_wave, summed_responses):
        maxidx_sta     = np.argmax(np.abs(sta))
        theta_led      = dome_positions[maxidx_sta//237,maxidx_sta%237,1]
        phi_led        = dome_positions[maxidx_sta//237,maxidx_sta%237,2]
        relative_waves = get_waves_relative_position((theta_led, phi_led), mode="spherical")

        stas_position_l.append((theta_led, phi_led))
        waves_position_l.append(relative_waves)

        waves_distance = relative_waves[:,1]
        waves_angle    = (relative_waves[:,2]+tau)%(tau) #Set the angle in the (0,2pi) range

        looming_mask   = (waves_distance<looming_distance_th)
        shrink_mask    = (waves_distance>np.pi-looming_distance_th)
        waves_mask     = (waves_distance>moving_distance_th) & (waves_distance<np.pi-moving_distance_th)

        vectors_dir  = np.exp(waves_angle*1j)  #Create vectors using imaginary numbers
        vectors_ori  = np.exp(waves_angle*1j*2)#x2 gather the vectors with opposite directions
        dir_pref     = np.nan_to_num((vectors_dir[waves_mask] * cell_responses[waves_mask]).sum() / cell_responses[waves_mask].sum())
        ori_pref     = np.nan_to_num((vectors_ori[waves_mask] * cell_responses[waves_mask]).sum() / cell_responses[waves_mask].sum())
        ds_idx       = abs(dir_pref)
        os_idx       = abs(ori_pref)

        looming_response   = (cell_responses[looming_mask]).sum()
        shrinking_response = (cell_responses[shrink_mask]).sum()
        looming_idx        = (looming_response-shrinking_response)/(looming_response+shrinking_response)

        ori_pref_l.append(ori_pref); dir_pref_l.append(dir_pref)
        ori_idx_l.append(os_idx); dir_idx_l.append(ds_idx)
        loom_idx_l.append(looming_idx)

        np.random.seed(1)
        rand_ori_idx_l  = np.empty(n_bootstrap)
        rand_dir_idx_l  = np.empty(n_bootstrap)
        rand_loom_idx_l = np.empty(n_bootstrap)
        for i in range(n_bootstrap):
            shuffled_response = cell_responses.copy()
            np.random.shuffle(shuffled_response)

            rand_dir_pref     = np.nan_to_num((vectors_dir[waves_mask] * shuffled_response[waves_mask]).sum() / shuffled_response[waves_mask].sum())
            rand_ori_pref     = np.nan_to_num((vectors_ori[waves_mask] * shuffled_response[waves_mask]).sum() / shuffled_response[waves_mask].sum())
            rand_dir_idx_l[i] = abs(rand_dir_pref)
            rand_ori_idx_l[i] = abs(rand_ori_pref)

            rand_looming_response   = (shuffled_response[looming_mask]).sum()
            rand_shrinking_response = (shuffled_response[shrink_mask]).sum()
            rand_loom_idx_l[i]      = (rand_looming_response-rand_shrinking_response)/(rand_looming_response+rand_shrinking_response)

        #Same calculation of pval as in Baden et al 2016
        p_val_dir  = 1 - (np.sum(rand_dir_idx_l<ds_idx)/n_bootstrap)
        p_val_ori  = 1 - (np.sum(rand_ori_idx_l<os_idx)/n_bootstrap)
        p_val_loom = 1 - (np.sum(np.abs(rand_loom_idx_l)<abs(looming_idx))/n_bootstrap)

        ori_pval_l.append(p_val_ori); dir_pval_l.append(p_val_dir); loom_pval_l.append(p_val_loom)

        # original orientation, by divinding the phase of the vector by two
        polar_ori_pref = polar(ori_pref)
        new_vector     = ((polar_ori_pref[1]+tau)%tau)/2 #Convert to positive radian angle and divide by two
        ori_pref       = rect(polar_ori_pref[0], new_vector)

    return (summed_responses,
            dir_pref_l, dir_idx_l, dir_pval_l,
            ori_pref_l, ori_idx_l, ori_pval_l,
            loom_idx_l, loom_pval_l,
            stas_position_l, waves_position_l)

# Cell
def peri_saccadic_response(spike_counts, eye_track, motion_threshold=5, window=15):
    """


    Computes the cell average response around saccades.

    params:
        - spike_counts: cells activity matrix of shape (t, n_cell)
        - eye_track: Eye tracking data of shape (t, x_pos, y_pos, ...)
        - motion_threshold: Amount of motion in pixel to account for a saccade
        - window: Size of the window before and after the saccade on which to average the cell response

    return:
        - peri saccadic response of cells of shape (n_cell, window*2+1)
    """
    eye_shifts = np.concatenate(([0],
                                 np.linalg.norm(eye_tracking[1:,:2]-eye_tracking[:-1,:2], axis=1)))

    #Because eye tracking is usually upsampled from 15 to 60Hz, it sums the shift, and smooth the peak
    # detection
    summed_shifts = np.convolve(eye_shifts, [1,1,1,1,1,1,1], mode="same")

    peaks, res = signal.find_peaks(eye_shifts, height=motion_threshold, distance=10)
    heights = res["peak_heights"] #Not used for now

    psr = np.zeros((window*2, spike_bins.shape[1]))
    for peak in peaks:
        if peak<window or (peak+window)>len(spike_bins):
            continue #Just ignoring peaks too close to the matrix edges
        psr += spike_bins[peak-window:peak+window]
    psr /= len(peaks)
    return psr


# AUTOGENERATED! DO NOT EDIT! File to edit: 03_modelling.ipynb (unless otherwise specified).

__all__ = ['sigmoid', 'gaussian', 'sum_of_gaussian', 'diff_of_gaussian', 'gaussian_2D', 'sum_of_2D_gaussian',
           'exponential_decay', 'sin_exponent', 'sinexp_sigm', 'fit_sigmoid', 'fit_spatial_sta', 'fit_temporal_sta',
           'fit_chirp_am', 'fit_chirp_freq_epoch', 'fit_transiency', 'repetition_quality_index',
           'onoff_transient_index']

# Cell
import math
import scipy as sp
import scipy.optimize
from scipy.interpolate import interp1d
import numpy as np
from functools import partial

# Cell
_default_sigmoid_p = (1,1,0,0)
def sigmoid(x, sigma, amp, x0, y0):
    """Sigmoid function
    params:
        - x: 1D numpy array at which to evaluate the points
        - sigma: steepness of the sigmoid
        - amp: amplitude of the sigmoid
        - x0: shift in x of the sigmoid
        - y0: shift in y of the sigmoid
    """
    return amp / (1.0 + np.exp(-sigma*(x-x0))) + y0

# Cell
_default_gaussian_p = (1,1,0,0)
def gaussian(x, sigma, amp, x0, y0):
    """Gaussian function
    params:
        - x: 1D numpy array at which to evaluate the points
        - sigma: width of the gaussian
        - amp: amplitude of the gaussian
        - x0: shift in x of the gaussian
        - y0: shift in y of the gaussian
    """
    return amp * np.exp(-(x-x0)**2 / (2*sigma**2)) + y0

# Cell
def sum_of_gaussian(t, sigma_1, amp_1, x0_1,
                       sigma_2, amp_2, x0_2, y0):
    """
    Sum of gaussian, using `gaussian`
    params:
        - t: 1D numpy array at which to evaluate the points
        - sigma_1: width of the 1st gaussian
        - amp_1: amplitude of the 1st gaussian
        - x0_1: shift in x of the 1st gaussian
        - sigma_2: width of the 2nd gaussian
        - amp_2: amplitude of the 2nd gaussian
        - x0_2: shift in x of the 2nd gaussian
        - y0: shift in y of the gaussian (shared for the two gaussians)"""
    return (gaussian(t, sigma_1, amp_1, x0_1, 0)
            + gaussian(t, sigma_2, amp_2, x0_2, 0) + y0)

def diff_of_gaussian(t, sigma_1, amp_1, x0_1,
                       sigma_2, amp_2, x0_2, y0):
    """
    Difference of gaussian, using `gaussian`
    params:
        - t: 1D numpy array at which to evaluate the points
        - sigma_1: width of the 1st gaussian
        - amp_1: amplitude of the 1st gaussian
        - x0_1: shift in x of the 1st gaussian
        - sigma_2: width of the 2nd gaussian
        - amp_2: amplitude of the 2nd gaussian
        - x0_2: shift in x of the 2nd gaussian
        - y0: shift in y of the gaussian (shared for the two gaussians)"""
    return (gaussian(t, sigma_1, amp_1, x0_1, 0)
            - gaussian(t, sigma_2, amp_2, x0_2, 0) ) + y0

# Cell
_default_2Dgaussian_p = (1,1,1,0,0,0,0)
def gaussian_2D(xz, sigma_x, sigma_z, amp, theta, x0, z0, y0):
    """Two dimensional Gaussian function
    params:
        - xz: meshgrid of x and z coordinates at which to evaluate the points
        - sigma_x: width of the gaussian
        - sigma_z: height of the gaussian
        - amp: amplitude of the gaussian
        - theta: angle of the gaussian (in radian)
        - x0: shift in x of the gaussian
        - z0: shift in z of the gaussian
        - y0: shift in y of the gaussian
    """
    (x,z) = xz
    x0, z0 = float(x0), float(z0)
    a =  (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_z**2)
    b = -(np.sin(2*theta)) /(4*sigma_x**2) + (np.sin(2*theta)) /(4*sigma_z**2)
    c =  (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_z**2)
    g = amp * np.exp( -(a*((x-x0)**2) + 2*b*(x-x0)*(z-z0) + c*((z-z0)**2))) + y0
    return g.ravel()

# Cell
def sum_of_2D_gaussian(xz, sigma_x_1, sigma_z_1, amp_1, theta_1, x0_1, z0_1,
                           sigma_x_2, sigma_z_2, amp_2, theta_2, x0_2, z0_2, y0):
    """Sum of two 2D Gaussian function. For the params, see `gaussian_2D`.
     However, both share the y0 parameter."""
    return (gaussian_2D(xz, sigma_x_1, sigma_z_1, amp_1, theta_1, x0_1, z0_1, 0)
            + gaussian_2D(xz, sigma_x_2, sigma_z_2, amp_2, theta_2, x0_2, z0_2, 0) + y0)

# Cell
def exponential_decay(x, tau, baseline, amplitude):
    """Exponential decay function
    params:
        - x: 1D numpy array at which to evaluate the points
        - tau: decay speed
        - baseline: baseline after the decay. Must be lower than the amplitude
        - amplitude: Initial amplitude before the decay
    """
    return np.exp(-x*tau)*(amplitude-baseline) + baseline

# Cell
def sin_exponent(x, amp, phi, freq, exp):
    """Sine raised to an exponent power. To the power 2, the sine is going between 0 and 1
    while its period is halfed. Subsequent round powers have the effect of narowing the peaks
    and making the rest of the sine approach zero. (graph it to see it)
    params:
        - x: 1D numpy array at which to evaluate the points
        - amp: amplitude of the sine
        - phi: phase of the sine in radian
        - freq: frequency of the sine in Hz
        - exp: Power the sine is raised to"""

    #Multiply by pi instead of 2*pi because it expect the cos to be at least squared
    #Same goes for phi that I halfed so the period is still 2*pi
    return np.power(np.sin((x*freq*np.pi)+(phi/2)), exp)*amp


# Cell
def sinexp_sigm(x, sigma, x0, y0, amp, phi, freq, exp):
    """A sine exponent weighted by a sigmoid. For parameters, see `sigmoid` and `sin_exponent`.
    However the amp and y0 (baseline) parameters are shared between the two sub-functions"""
    return sin_exponent(x, amp, phi, freq, exp) * sigmoid(x, sigma, 1, x0, 0) + y0

# Cell
def fit_sigmoid(nonlin, t=None):
    """Fit a sigmoid to a 1D input. Used to fit the nonlinearities of the cells responses.
    Also, the data must be increasing (from left to right).
    params:
        - nonlin: 1D data to fit
        - t     : timepoints of the data to fit. Same lenght as nonlin

    return:
        - A dictionary containing the parameters of the fit for a `sigmoid`
        - A quality index of the fit (explained variance of the model)"""

    assert np.mean(nonlin[:20])<np.mean(nonlin[-20:]), "Nonlinearity must be increasing from left to right"
    if t is None:
        t = np.arange(len(nonlin))
    bounds = ([0.000001, 0.000001, -np.inf, 0],
              [np.inf  , np.inf,    np.inf, np.max(nonlin)])
    try:
        fit, _ = sp.optimize.curve_fit(sigmoid, t, nonlin, maxfev=1000000, bounds=bounds)
        fit  = {"sigma":fit[0],"amp":fit[1],"x0":fit[2],"y0":fit[3]}
    except RuntimeError:
        fit = {"sigma":1,"amp":0,"x0":0,"y0":0}
    model = sigmoid(t, **fit)
    quality_index =  1 - (np.var(nonlin-model)/np.var(nonlin))
    return fit, quality_index

def fit_spatial_sta(sta):
    """Fit a sum of 2D gaussian to a 2D matrix. Used to fit the STA obtained from checkerboard like stimulus.
    params:
        - sta: 2D data to fit

    return:
        - A dictionary containing the parameters of the fit for a `sum_of_2D_gaussian`
        - A quality index of the fit (explained variance of the model)"""
    shape_y, shape_x = sta.shape
    eps = 0.000001
    argmax = np.unravel_index(sta.argmax(), sta.shape)
    argmin = np.unravel_index(sta.argmin(), sta.shape)
    #                   sigma_x    sigma_z   amp theta      x0        z0      y0
    bounds_center   = ((0,         0,         0,  0,         0,       0,       0),
                       (shape_x*4, shape_y*4, 2,  2*math.pi, shape_x, shape_y, eps))
    if sta[argmax] < abs(sta[argmin]):
        argmax, argmin  = argmin, argmax
        bounds_center   = ((0,         0,        -2,  0,         0,       0,       0),
                           (shape_x*4, shape_y*4, 0,  2*math.pi, shape_x, shape_y, eps))

    x    = np.arange(shape_x)
    z    = np.arange(shape_y)
    x, z = np.meshgrid(x, z)
    #                   sigma_x sigma_z  amp        theta      x0        z0       y0
    init_fit_center   = (  1,     1,    sta[argmax],  0,   argmax[1],  argmax[0], 0)
    data_tofit          = sta.reshape(-1)
    if np.isnan(sp.sum(sta)): #We check that the sta exists, otherwise return default zero model
        fit  = {"sigma_x_1": 1, "sigma_z_1":1, "amp_1":0, "theta_1":0, "x0_1":0, "z0_1":0,
                "sigma_x_2": 1, "sigma_z_2":1, "amp_2":0, "theta_2":0, "x0_2":0, "z0_2":0, "y0":0}
    else:
        try:
            fit_c, _   = sp.optimize.curve_fit(gaussian_2D, (x,z), data_tofit, p0=init_fit_center, bounds=bounds_center)
            tmp_to_fit =  data_tofit - gaussian_2D((x,z), *fit_c)
            #                   sigma_x    sigma_z   amp theta        x0            z0           y0
            bounds_surround = ((0,         0,        -2,  0,         fit_c[4],     fit_c[5],     0),
                               (shape_x*4, shape_y*4, 0,  2*math.pi, fit_c[4]+eps, fit_c[5]+eps, eps))
            init_fit_surround = (  2,      2,         0,  0,         fit_c[4],     fit_c[5],     0)
            fit_s, _   = sp.optimize.curve_fit(gaussian_2D, (x,z), tmp_to_fit, p0=init_fit_surround, bounds=bounds_surround)

            init_fit_sum = (*fit_c[:-1],*fit_s[:-1],0)
            #Bounding only the position, that center fit should have found
            bounds_sum   = ((-np.inf,-np.inf,-np.inf,-np.inf,fit_c[4],fit_c[5],
                             -np.inf,-np.inf,-np.inf,-np.inf,fit_c[4],fit_c[5],-np.inf),
                            (np.inf, np.inf, np.inf, np.inf, fit_s[4]+eps,fit_s[5]+eps,
                             np.inf,np.inf,np.inf,np.inf,fit_s[4]+eps,fit_s[5]+eps,np.inf))
            fit_sum, _ = sp.optimize.curve_fit(sum_of_2D_gaussian, (x,z), data_tofit, p0=init_fit_sum, bounds=bounds_sum)
            fit  = {"sigma_x_1": fit_sum[0], "sigma_z_1":fit_sum[1], "amp_1":fit_sum[2],
                    "theta_1":fit_sum[3], "x0_1":fit_sum[4], "z0_1":fit_sum[5],
                    "sigma_x_2": fit_sum[6], "sigma_z_2":fit_sum[7], "amp_2":fit_sum[8],
                    "theta_2":fit_sum[9], "x0_2":fit_sum[10], "z0_2":fit_sum[11], "y0":fit_sum[12]}

        except RuntimeError: #If a model can't be fitted, we get a runtimeError
            fit  = {"sigma_x_1": 1, "sigma_z_1":1, "amp_1":0, "theta_1":0, "x0_1":0, "z0_1":0,
                    "sigma_x_2": 1, "sigma_z_2":1, "amp_2":0, "theta_2":0, "x0_2":0, "z0_2":0, "y0":0}
    model = sum_of_2D_gaussian((x,z), **fit)
    quality_index = 1 - (np.var(data_tofit-model)/np.var(data_tofit))
    return fit, quality_index


# Cell
def fit_temporal_sta(sta, frame_rate=60):
    """Fit a sum of gaussian to a 1D input. Used to fit the STA obtained from fullfield flicker like stimulus.
    params:
        - sta: 1D data to fit

    return:
        - A dictionary containing the parameters of the fit for a `diff_of_gaussian`. The parameter with _1 suffix
        correspond to the closest gaussian to the zero timepoint.
        - A quality index of the fit (explained variance of the model)"""

    # Cutting out the tail of the tSTA to increase the fitting error of the "interesting" part
    cell_abs = np.abs(sta)
    startSTA   = len(sta)-np.argmax(cell_abs>(cell_abs.max()*.4))
    startSTA  += np.maximum(5,(startSTA*.5)).astype(int)
    fullsta = sta.copy()
    sta = sta[-startSTA:]

    t = np.linspace((1-len(sta))/frame_rate, 0, len(sta))

                #sigma_1, amp_1,   x0_1,                sigma_2, amp_2,  x0_2,
    bounds   = ((0,       0,      -len(sta)/frame_rate, 0,       0,      -len(sta)/frame_rate),
                (.1,      np.inf, 1/frame_rate,        .1,       np.inf, 1/frame_rate))

    diff_of_gauss = partial(diff_of_gaussian, y0=0)

    if np.isnan(sp.sum(sta)): #We check that the sta exists, otherwise return default zero model
        fit  = {"sigma_1":1,"amp_1":0,"x0_1":0,
                "sigma_2":1,"amp_2":0,"x0_2":0,"y0":0}
    else:
        try:
            fit, _ = sp.optimize.curve_fit(diff_of_gauss, t, sta, bounds=bounds, max_nfev=100000000000)

            fit  = {"sigma_1":fit[0],"amp_1":fit[1],"x0_1":fit[2],
                    "sigma_2":fit[3],"amp_2":fit[4],"x0_2":fit[5],"y0":0}
        except RuntimeError: #If a model can't be fitted, we get a runtimeError
            fit  = {"sigma_1":1,"amp_1":0,"x0_1":0,
                    "sigma_2":1,"amp_2":0,"x0_2":0,"y0":0}

    t = np.linspace((1-len(fullsta))/frame_rate, 0, len(fullsta))

    model = diff_of_gaussian(t, **fit)
    quality_index = 1 - (np.var(fullsta-model)/np.var(fullsta))
    return fit, quality_index


# Cell
def fit_chirp_am(cell_mean, start=420, stop=960, freq=1.5, frame_rate=60):
    """Fit a sinexp_sigm to the mean response of a cell to chirp_am stimulus.
    params:
        - cell_mean: Average response of the cell to the chirp_am stimulus
        - start:     Index in the cell mean where to start the fitting
        - stop:      Index in the cell mean where to stop the fitting
        - freq:      Frequency of the amplitude modulation of the stimulus in Hz

    return :
        - A dictionary containing the parameters of the fit for a `sinexp_sigm`
        - A quality index of the fit (explained variance of the model)"""

    to_fit = cell_mean[start:stop]
    t = np.linspace(0, len(to_fit)/frame_rate, len(to_fit), endpoint=False)

    #If suppressed by contrast cell, the firing should be higher in first part of chirp
    fit_positive = np.mean(to_fit[:len(to_fit)//3]) < np.mean(to_fit[len(to_fit)*2//3:])

    tmp_diff = np.inf
    best_fit = (1, 0, 0, 0, 0, freq, 0)
    phi      = np.pi
    for exp in np.exp2(range(1,10)): #Fitting the data with different sin exponents, to narrow the fit
        try:
            sinexp_sigm_part = partial(sinexp_sigm, freq=freq, exp=exp)
            if fit_positive:
                fit, _ = sp.optimize.curve_fit(sinexp_sigm_part, t, to_fit,
                                               bounds=[(  0,    0          ,        0       ,      0        , 0),
                                                        (50, len(to_fit)/60,  np.max(to_fit), np.max(to_fit), 2*np.pi)],
                                               p0 =      (1, len(to_fit)/120,       0       , np.max(to_fit), phi),
                                              maxfev=100000)
            else:
                fit, _ = sp.optimize.curve_fit(sinexp_sigm_part, t, to_fit,
                                               bounds=[(  0,    0          ,        0       ,-np.max(to_fit), 0),
                                                        (50, len(to_fit)/60,  np.max(to_fit),     0         , 2*np.pi)],
                                               p0 =     (1, len(to_fit)/120,  np.max(to_fit) ,-np.max(to_fit),phi),
                                              maxfev=100000)

            mse = np.sum(np.square(sinexp_sigm_part(t, *fit) - to_fit))
            phi = fit[4]
            if mse < tmp_diff:
                best_fit = (*fit, freq, exp)
                tmp_diff = mse
        except:
            continue
    best_fit = dict((k, v) for v, k in zip(best_fit, ["sigma","x0","y0","amp","phi","freq","exp"]))

    model  = sinexp_sigm(t, **best_fit)
    to_fit = cell_mean[start:stop]
    quality_index = 1 - (np.var(to_fit-model)/np.var(to_fit))
    return best_fit, quality_index

def fit_chirp_freq_epoch(cell_mean, start=360, freqs=[1.875,3.75,7.5,15,30], durations=[2,2,2,1,1], frame_rate=60):
    """Fit multiple sinexp_sigm to the mean response of a cell to chirp_freq_epoch stimulus.
    Each epoch is fitted independantly by a `sin_exponent`
    params:
        - cell_mean: Average response of the cell to the chirp_am stimulus
        - start:     Index in the cell mean where to start the fitting
        - freqs:     Frequencies of each epoch in Hz
        - durations: Duration of each epoch in seconds
        - freq:      Frequency of the amplitude modulation of the stimulus

    return :
        - List of dictionary containing the parameters of each epoch fit for a `sin_exponent`
        - List of quality index of each epoch fit (explained variance of the model)"""
    best_fit_l = []
    qualityidx_l = []

    cursor = start #Start of the freqs

    #We fit each freqency epoch with a cos function
    for freq, dur in zip(freqs, durations):
        n_rep, n_fr = int(dur*freq), int(frame_rate/freq)
        edgecut = 10
        cursor += edgecut
        len_fit = n_rep*n_fr - edgecut

        #Upsampling of the data to prevent low sampling effects with high freq sine
        upsamp = 10
        to_fit = cell_mean[cursor:cursor+len_fit+1]
        t = np.linspace(0, len_fit/frame_rate, len_fit*upsamp, endpoint=False)
        to_fit = interp1d(np.linspace(0,len(to_fit)/frame_rate, len(to_fit), endpoint=False), to_fit)(t)

        best_fit = None
        tmp_diff = np.inf
        for exp in np.exp2(range(1,10)): #Fitting the data with different cos exponents, to narrow the fit
            try:
                sin_exponent_part = partial(sin_exponent, freq=freq, exp=exp)
                fit, cov = sp.optimize.curve_fit(sin_exponent_part, t, to_fit, bounds=[(0,          0),
                                                                                       (max(to_fit),np.pi*2)])
                mse = np.sum(np.square(sin_exponent_part(t, *fit) - to_fit))
                if mse < tmp_diff:
                    best_cov = cov
                    best_fit = (*fit, freq, exp)
                    tmp_diff = mse
            except:
                continue
        if best_fit is not None:
            best_fit = dict((k, v) for v, k in zip(best_fit, ["amp","phi","freq","exp"]))
            best_fit_l.append(best_fit)
            model = sin_exponent(t, **best_fit)
            qualityidx_l.append(1 - (np.var(to_fit-model)/np.var(to_fit)))
        else:
            best_fit_l.append({"amp":0,"phi":0,"freq":freq,"exp":0})
            qualityidx_l.append(0)

        cursor += len_fit
    return best_fit_l, qualityidx_l #, best_cov_l

# Cell
def fit_transiency(pref_response, frame_rate=60):
    """Fit an exponential decay to a 1D input. Used to fit the response of cells to their preferred
    sustained stimulus.
    params:
        - pref_response: 1D data to fit

    return:
        - A dictionary containing the parameters of the fit for a `exponential_decay`
        - A quality index of the fit (explained variance of the model)"""
    peak_position = np.argmax(pref_response)
    pref_response = pref_response[peak_position:]

    t = np.linspace(0, len(pref_response)/frame_rate, len(pref_response), endpoint=False)
    bounds = ([0.000001, 0],
              [np.inf  , pref_response[0]])
    if np.all(pref_response==0):
        fit = {"tau":1,"baseline":0,"amplitude":0}
    else:
        try:
            decay_part = partial(exponential_decay, amplitude=pref_response[0])
            fit, _ = sp.optimize.curve_fit(decay_part, t, pref_response, maxfev=1000000, bounds=bounds)
            fit  = {"tau":fit[0],"baseline":fit[1],"amplitude":pref_response[0]}
        except RuntimeError:
            fit = {"tau":1,"baseline":0,"amplitude":0}
    model = exponential_decay(t, **fit)
    quality_index =  1 - (np.var(pref_response-model)/np.var(pref_response))

    return fit, quality_index, peak_position

# Cell
def repetition_quality_index(cell_response):
    """Return a quality index of cell response to a repeated stimulus.
    params:
        - cell_response: response of a cell of shape (n_rep, time)"""
    tmp1 = np.var(np.mean(cell_response, axis=0)) #VAR(MEAN(C)repeat)time
    tmp2 = np.mean(np.var(cell_response, axis=0)) #MEAN(VAR(C)time)repeat
    quality = tmp1/tmp2
    return quality

def onoff_transient_index(cell_response, start_on=120, stop_on=240, start_off=240, stop_off=360):
    """Return both on-off and transient indexes of cell response. Transiency index is not always
    ideal to describe the cells response. Consider fitting an exponential decay with `fit_transiency`.
    params:
        - cell_response: response of a cell of shape (time,...)
        - start_on:  starting index of ON stimulation
        - stop_on:   stop index of ON stimulation
        - start_off: starting index of OFF stimulation
        - stop_off:  stop index of OFF stimulation
    return:
        - ON-OFF index
        - Transiency index"""
    #ON-OFF index
    on_response   = cell_response[start_on:stop_on]
    on_sum        = np.sum(on_response)
    off_response  = cell_response[start_off:stop_off]
    off_sum       = np.sum(off_response)
    onoff_indexes = (on_sum-off_sum)/(on_sum+off_sum)

    #Transient index
    if onoff_indexes>0: #Main ON response
        transient_sum   = np.sum(cell_response[start_on:start_on+30])
        sustained_sum   = np.sum(cell_response[start_on:stop_on])
    else:
        transient_sum   = np.sum(cell_response[start_off:start_off+30])
        sustained_sum   = np.sum(cell_response[start_off:stop_off])
    transient_index = transient_sum/sustained_sum

    return onoff_indexes, transient_index


# AUTOGENERATED! DO NOT EDIT! File to edit: 06_eyetrack.ipynb (unless otherwise specified).

__all__ = ['interpolate_screen_pos', 'interpolate_checker_pos', 'split_eye_events', 'get_spherical_map',
           'apply_spherical_map']

# Cell
import numpy as np
import scipy.interpolate as interpolate
from sklearn import cluster

# Cell
def interpolate_screen_pos(screen_pos, xnew, ynew, kind='linear'):
    """
    Interpolate the position of the xnew and ynew pixels from the original screen_pos.
    `interpolate_checker_pos` should be used instead as it's more user friendly.
    params:
        - screen_pos: Screen positions in shape (17, 10, 2) obtained from calibration (from 80x80 pixel checker corners on a 1280x720px screen)
        - xnew: New pixels indexes in x, in interval [0, 17[
        - ynew: New pixels indexes in y, in interval [0, 10[
    return:
        - Screen positions in shape (len(xnew),len(ynew), 2)
    """
    f = interpolate.interp2d(np.arange(17), np.arange(10), screen_pos[:,:,0].T, kind=kind)
    znew_x = f(xnew, ynew)

    f = interpolate.interp2d(np.arange(17), np.arange(10), screen_pos[:,:,1].T, kind=kind)
    znew_y = f(xnew, ynew)

    return np.stack((znew_x, znew_y), axis=-1)

def interpolate_checker_pos(screen_pos, width_box, height_box, kind='linear'):
    """
    Interpolate the centre of the checkerboard pixels from the screen calibrated position.
    params:
        - screen_pos: Screen positions in shape (17, 10, 2) obtained from calibration (from 80x80 pixel checker corners on a 1280x720px screen)
        - width_box: Width in pixel of a box
        - height_box: Height in pixel of a box
        - kind: kind of interpolation in {'linear', 'cubic', 'quintic'}
    """
    assert 1280%width_box==0, "unpredictable behaviour if 1280 is not a multiple of width_box"
    assert 720%height_box==0, "unpredictable behaviour if 720 is not a multiple of height_box"
    n_x = 1280/width_box
    n_y = 720/height_box
    xnew = np.arange(16/n_x/2, 16+16/n_x/2, 16/n_x)
    ynew = np.arange(9/n_y/2, 9+9/n_y/2, 9/n_y)
    return interpolate_screen_pos(screen_pos, xnew, ynew, kind=kind)

# Cell
def split_eye_events(eye_tracking, eps=2):
    """
    Split the record where the eye moves. Detection done with clustering on X,Y and time of the eye position.

    params:
        - eye_tracking: Eye traking array of the ellipse fit, in shape (t, (x,y,width,height,angle))
        - eps: Distance to detect eye movements. Adjust this parameter if results are not satisfying
        - kind: kind of interpolation in {'linear', 'cubic', 'quintic'}
    return:
        - move_indexes, blink_indexes, noise_indexes
    """
    x_pos    = np.array(eye_tracking[:,0])

    X        = np.stack((x_pos, np.linspace(0, len(x_pos), len(x_pos))*.5)).T
    clusters = cluster.dbscan(X, eps=eps, min_samples=5, metric='minkowski', p=2)
    move_indexes = np.where(clusters[1][1:] > clusters[1][:-1])[0] + 1

    noise_indexes = np.where(clusters[1] == -1)[0]
    blink_indexes = np.where(x_pos == 0)[0]

    return move_indexes, blink_indexes, noise_indexes

# Cell
def get_spherical_map(screen_pos, input_res=(281, 500), output_res=(360, 640), k_side=2, filling_pol="nonzero"):
    """
    Generate the mapping from normal image to an image projected on a spherical screen
    params:
        - screen_pos: Screen positions in shape (17, 10, (elev, azim)) obtained from calibration (from 80x80 pixel checker corners on a 1280x720px screen)
        - input_res: resolution of the input image
        - output_res: resolution of the output image
        - k_side: Kernel side's size to fill holes in the mapped image (2 -> 5*5 kernel)
        - filling_pol: Filling policy for the empty values, to which `f_fill` of `apply_spherical_map` will be applied
            nonzero: indexes of all nonzero elements
            closest: index of the closest nonzero element

    returns:
        - A mapping to be used in `apply_spherical_map`
    """
    assert filling_pol in ["nonzero", "closest"]

    screen_interp = interpolate_screen_pos(screen_pos, np.linspace(0,16, input_res[1], endpoint=True),
                                                       np.linspace(0, 9, input_res[0], endpoint=True))
    y_inres, x_inres = input_res
    y_res, x_res     = output_res
    xnew = np.linspace(screen_interp[:,:,1].min(), screen_interp[:,:,1].max(), x_res)
    ynew = np.linspace(screen_interp[:,:,0].min(), screen_interp[:,:,0].max(), y_res)
    map_img = np.zeros((y_res, x_res))

    y_map, x_map = np.empty(y_inres*x_inres, dtype=int), np.empty(y_inres*x_inres, dtype=int)
    for i, (y, x) in enumerate(zip(screen_interp[:,::-1,0].flatten(), screen_interp[:,::-1,1].flatten())):
        y_map[i] = np.argmin(ynew<y)
        x_map[i] = np.argmin(xnew<x)
    map_img[y_map, x_map] = 1

    y_nonzero, x_nonzero = np.nonzero(map_img==0) #Finds where the image is still zero
    fill_x_l, fill_y_l, nonzeros_l = [], [], []
    for y, x in zip(y_nonzero, x_nonzero):
        # Sets the limits to where to look for nonzeros pixels
        ylow, xlow = max(0, y-k_side), max(0, x-k_side)
        yhig, xhig = min(y+k_side+1, y_res), min(x+k_side+1, x_res)
        area = map_img[ylow:yhig, xlow:xhig]

        if np.any(area): #If there are pixels around
            fill_x_l.append(x)
            fill_y_l.append(y)
            nonz_y, nonz_x = np.nonzero(area)
            if filling_pol=="nonzero":
                nonzeros_l.append((nonz_y+ylow, nonz_x+xlow)) #store the nonzero slicing for later filling
            elif filling_pol=="closest":
                xx, yy = np.meshgrid(np.arange(xlow,xhig), np.arange(ylow,yhig))
                distances = np.sqrt((yy-y)**2+(xx-x)**2)
                idx_min   = np.argmin(distances[nonz_y, nonz_x])
                nonzeros_l.append(([nonz_y[idx_min]+ylow], [nonz_x[idx_min]+xlow]))

    return (y_map, x_map), (fill_y_l, fill_x_l, nonzeros_l)

def apply_spherical_map(img_src, mapping, output_res=(360, 640), f_fill=np.mean):
    """
    Apply a mapping to an input image
    params:
        - img_src: Source image to transform
        - mapping: Mapping obtained from `get_spherical_map`
        - output_res: Output resolution. Must match the parameter given to `get_spherical_map`
        - f_fill: Function to apply when filling the holes (e.g. np.median, np.mean)
    return:
        - The transformed image
    """
    (y_map, x_map), (fill_y_l, fill_x_l, nonzeros_l) = mapping

    transfo_img = np.zeros(output_res) + 128
    transfo_img[y_map, x_map] = img_src.flatten()

    for y, x, nonz in zip(fill_y_l, fill_x_l, nonzeros_l):
        transfo_img[y, x] = f_fill(transfo_img[nonz])
    return transfo_img


# AUTOGENERATED! DO NOT EDIT! File to edit: 99_testdata.ipynb (unless otherwise specified).

__all__ = ['load_vivo_2p']

# Cell
import numpy as np
import pickle
from os.path import join
import matplotlib.pyplot as plt

from .core import import_record


def load_vivo_2p(testdata_dir):
    """Use as locals().update(load_vivo_2p()) to load test calcium data (will change soon)
    """
    vivo_2p_dir = testdata_dir
    reM = import_record(join(vivo_2p_dir, "record_master.h5"))
    stim_d = {"0_darkness": np.load(join(vivo_2p_dir,"0_darkness.npy")),
             "38786_checkerboard": np.load(join(vivo_2p_dir,"38786_checkerboard.npy")),
             "82376_water": np.load(join(vivo_2p_dir,"82376_water.npy")),
             "126646_moving_gratings": np.load(join(vivo_2p_dir,"126646_moving_gratings.npy")),
             "174837_chirp_am": np.load(join(vivo_2p_dir,"174837_chirp_am.npy")),
             "195136_chirp_freq_epoch": np.load(join(vivo_2p_dir,"195136_chirp_freq_epoch.npy")),
             "213925_fullfield_flicker": np.load(join(vivo_2p_dir,"213925_fullfield_flicker.npy"))}
    S_matrix = np.load(join(vivo_2p_dir, "cells_spike_matrix.npy"))
    A_matrix = np.load(join(vivo_2p_dir, "cells_spatial_matrix.npy"))
    eye_TP   = np.load(join(vivo_2p_dir, "eyevid_frame_timepoints.npy"))
    eye_DATA = np.load(join(vivo_2p_dir, "eyevid_pupil_data.npy"))
    proj_TP  = np.load(join(vivo_2p_dir, "projector_frame_timepoints.npy"))
    proj_DATA= np.load(join(vivo_2p_dir, "projector_frame_data.npy"))
    treadm_DATA= np.load(join(vivo_2p_dir, "treadmill_data.npy"))
    len_records = np.load(join(vivo_2p_dir, "record_lengths.npy"))
    with open(join(vivo_2p_dir, "twoP_frame_timepoints.pkl"), mode="rb") as f:
        rec_TP = pickle.load(f)
    print("Returning stim_d, S_matrix, A_matrix, proj_TP, proj_DATA, eye_TP, eye_DATA, treadm_DATA, len_records, rec_TP, reM")
    return locals()